/*
 * epc - (e)eprom programmer controller version 1.0
 * Copyright (C) 1995 Phillip Musumeci.
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Program epc drives the Electronics Australia (E)EPROM Programmer
 * described in the September and October 1993 issues.  A PC with a
 * standard parallel port is required.
 *
 *
 * SOME CURRENT ASSUMPTIONS.
 *
 * 1) the programmer is connected to the parallel port at address 0x378.
 * 2) devices of capacity <= 16K are handled (change both symbols MAXMS and
 *    MAX_K_SIZE if you are increasing this).
 *
 *
 * COMPILER DEPENDENCIES.
 *
 * 1) low level port input input and output is handled by the functions:
 *      _inb( PORT );                  returns current value on port PORT
 *      _outb( value, PORT );          writes value to port PORT
 * 2) in-line assembly is used to read the current time with milli-second
 *    accuracy in function rtime() - this function may be compiler dependent
 *    (or already available in your C language support libraries).
 * 3) note the IO related define(s) below.
 */

#define FGETC_EOF -1 /* would normally be EOF */

/*************************************************************************/

char *pname[] = { "\n",
  "\n epc version 1.0, Copyright (C) 1995 Phillip Musumeci.",
  "\n epc comes with ABSOLUTELY NO WARRANTY; for details read file warrant.",
  "\n This is free software,  and  you are welcome to redistribute it under",
  "\n certain conditions; for details read file distrib.",
  "." };

char *whoami= "WORK:\n\
 Phillip Musumeci, Comp. Sys. Eng. Dept. RMIT, GPO Box 2476V, Melbourne 3001.";

char *email = "\n Email: phillip@cse.rmit.edu.au";

char *smith = "\n\
This software is free but if you wish to make a contribution,\n\
please send a (tax deductable) donation to a charity such as:\n\
    THE SMITH FAMILY, P.O. BOX 1000, WODEN. A.C.T. 2606.\n";

char *themag= "\
 epc drives the Electronics Australia (E)EPROM Programmer\n\
 described in the September and October 1993 issues.";

char *theact= "\
 epc version 1.0 kept pm active during summer break 1994/1995.";

/*************************************************************************/
#include <stdio.h>

#define BPL 32         /* = data bytes/line in output S-record HEX files */

#define MAXMS 16384    /* = size of HEX data buffers                  */
#define MAX_K_SIZE 16  /* = log_2( MAXMS / 1024 ) = number of K bytes */

#define EMPTY 0xff     /* = contents of an erased ROM location */

#define MASK8 0xff     /* = AND mask to get byte */

char a[MAXMS];         /* dual HEX data buffers */
char b[MAXMS];

unsigned i,j,k;        /* general counters and indices */
unsigned addr,mem_size;
int k_size;            /* device size in K bytes       */

int lookup[256];
/*
 * External 8bit data is   read into the  PC  by performing two  4bit  wide
 * reads.  The 4 data lines used are arbitrarily located (from our point of
 * view)   and some  are  inverted ---  this   lookup  table  speeds up the
 * decoding.  An alternative method  of decoding the 4bit  data would be to
 * use bit manipulation.
 */

#define DOUT 0x378
#define DIN  0x379     /* IO port addresses     */
#define CTRL 0x37A

char cmd,c;            /* temp char             */
char now[10];          /* holds current time    */

#define EPROM  0       /* define known chip_type values */
#define EEPROM 1
char *device_name[]={ "EPROM ","EEPROM" };
int chip_type;         /* store chip type */

/*************************************************************************/

/* define some external functions so we can use GNU C for code checking  */

extern int printf();
extern int fprintf();
extern int fclose();
extern int scanf();
extern int sscanf();
extern char getchar();
extern void times();

extern unsigned _inb();
extern void     _outb();

/*************************************************************************/

/*
 * Display message string arrays (such arrays are necessary if your
 * compiler cannot handle long strings).
 */

void print_array( msg )
char *msg[];
{
  int i;
  for( i=0; *msg[ i ] != '.' ; i++ )
    printf( msg[ i ] );
}

/*************************************************************************/

/*
 * Time delay support routines.
 */

                               /* (intended for PUBLIC access)         */
int tdel1,tdel2;               /* tdel1 and tdel2 specify time_delay() */
                               /* input parameters for a time delay    */

                               /* (intended for PRIVATE use)           */
int delta_t;                   /* interval in milliseconds             */
char tdata1[20];               /* time call storage                    */
char tdata2[20];               /* time call storage                    */
int tdel2_10,tdel2_50;         /* tdel2 values for 10ms,50ms delays    */

/*
 * time_delay() is the basic time delay function.  By using 2 nested loops,
 * it is  possible to adjust m  for calibration and  to  later adjust n for
 * scaling.  The  response to  scaling with  n is  not strictly linear  but
 * seems OK in practice.
 */

void time_delay( m,n )
  int m,n;
{
  int x,y;
  for( x=0 ; x<m ; x++ ){
    for( y=0 ; y<n ; y++ ){
    };
  };
}

/*
 * rtime() returns time-of-day of: hours, minutes, seconds, & milliseconds.
 */

#ifdef GNUC_TEST       /* define rtime external when code checking with GNU C */
extern void rtime();
#else
void rtime( x )
char *x;               /* pointer x supplies the address to store the result  */
{
#asm
        mov        ah,2ch        ; select time-of-day DOS call
        int        21h           ; perform DOS call with result in cx,dx
        mov        si,[bp+4]     ; copy address of answer (x) into register si
        mov        [si],cx       ; save cx result into *x and *(x+1)
        mov        [si+2],dx     ; save dx result into *(x+2) and *(x+3)
#
}
#endif

/*
 * measure_elapsed() executes a delay  specified by variables {tdel1,tdel2}
 * and returns the milli-second delay time in delta_t.
 */

void measure_elapsed()
{
  rtime( tdata1 ); time_delay( tdel1,tdel2 ); rtime( tdata2 );

                                        /* determine the elapsed time in ms  */
  delta_t = (int) *(tdata2+1) - (int) *(tdata1+1);
  if( (int) *(tdata1+1) > (int) *(tdata2+1) ) delta_t = 1;            /* hr  */
  delta_t = delta_t * 60 + ( (int) *(tdata2+0) - (int) *(tdata1+0) ); /* min */
  delta_t = delta_t * 60 + ( (int) *(tdata2+3) - (int) *(tdata1+3) ); /* sec */
                                                                      /* ms  */
  delta_t = delta_t * 1000 + ( (int) *(tdata2+2) - (int) *(tdata1+2) ) * 10;
}

/*
 * calibrate_timer() first measures the  time delay obtained from arbitrary
 * down count values and  then computes  parameters  for a 50ms  delay.  An
 * average result for a 50ms and a 10ms delay is displayed (an average test
 * is necessary as the time-of-day is available with 10ms accuracy).
 */

void calibrate_timer( verbose )
  int verbose;                 /* !=0 flags extra display  */
{
  printf("*****************************************************************\n");
  printf("TIMER CALIBRATION - please wait a few seconds ...");

  tdel1 = 10000;   /* test 1 - measure time delay for arbitrary settings */
  tdel2 = tdel2_50;

  measure_elapsed();           /* compute elapsed time in variable delta_t */

  if( verbose != 0 ){          /* verbose mode displays time data bytes */
    printf("\n Time stamps are: %2d:%2d:%2d.%2d", (int) *(tdata1+1), (int) *(tdata1+0), (int) *(tdata1+3), (int) *(tdata1+2));
    printf("\n                  %2d:%2d:%2d.%2d", (int) *(tdata2+1), (int) *(tdata2+0), (int) *(tdata2+3), (int) *(tdata2+2));
  }
  printf("\n %dx%d delay gives %d ms\n", tdel1,tdel2,delta_t);

                                               /* determine 50ms delay count */
  tdel1 = (int) ( (double) tdel1 * (double) 50 / (double) delta_t );

  printf("\
The delay routines will now be called 80 times so that an AVERAGE\n\
measure of time delay can be *estimated* from the 10ms resolution\n\
time-of-day DOS call available.\n");

                       /* test 2 - check 50ms delay */
  for( i=j=0; i<80; i++ ){
    measure_elapsed();         /* compute elapsed time in variable delta_t */
    j = j + delta_t;                           /* sum the delays achieved */
  }
  printf(" %dx%d delay gives %d ms\n", tdel1,tdel2,j/80);

                       /* test 3 - check 10ms delay */
  tdel2 = tdel2_10;
  for( i=j=0; i<80; i++ ){
    measure_elapsed();         /* compute elapsed time in variable delta_t */
    j = j + delta_t;                           /* sum the delays achieved */
  }
  printf(" %dx%d delay gives %d ms\n", tdel1,tdel2,j/80);

  tdel2 = tdel2_50;            /* restore settings for 50ms delay */

  printf("*****************************************************************\n");
}

/*************************************************************************/

/*
 * buffer() returns  a pointer to a buffer  chosen by  the user (address is
 * either buffer a or b).
 */

char *bmsg;            /* points to default buffer select message */

char *bmsg_a = "A/B";  /* bmsg_a and bmsg_b hold buffer choice messages */
char *bmsg_b = "B/A";  /* for A then B and B then A respectively.       */

char* buffer()
{
  char c;

  printf("%s ? ",bmsg);
  do {                          /* loop until suitable choice entered */
    c = getchar();
    switch( c ){
    case 'a': c='A';            /* specific entries */
      break;
    case 'b': c='B';
      break;
    default:  c = *bmsg;        /* accept default (1st char in bmsg) */
      break;
    }
  } while ( (c != 'A') & (c != 'B') );
  printf("\nBuffer = %c ",c);

  switch( c ){                  /* return buffer address          */
  case 'A': return a; break;    /* (include unreachable break)    */
  case 'B': return b; break;
  }
  return a;                     /* default return (never reached) */
}

/*************************************************************************/

/*
 * read_byte() reads the data  bus in the  PROM programmer.  As only a 4bit
 * input port is available, the 74HC157 chip  is switched to allow input of
 * low and high nibbles.  The array lookup[] provides high speed mapping of
 * input patterns into valid 4bit data.
 */

char read_byte()
{
  char l,h;

  _outb( 0x01, DOUT );          /* input LS 4 bits */
  l = lookup[ ((_inb( DIN ) | 0x07) & 0xBF) ];
  _outb( 0x00, DOUT );          /* input MS 4 bits */
  h = lookup[ ((_inb( DIN ) | 0x07) & 0xBF) ];

  return (h << 4 | l);          /* return 8 data bit result */
}

/*************************************************************************/

/*
 * select_device() returns the device memory  size in variables mem_size  &
 * k_size, sets  the  chip_type, and sets  the  programmer  timing variable
 * tdel2.
 */

void select_device()
{
                       /* input memory size */
  do {
    printf("\nSelect device size in K bytes (1");
    for( i=1; i<MAX_K_SIZE; ){                /* display possible answers */
      i = i*2;
      printf(",%d",i);
    }
    printf(") ? ");
    scanf("%d",&k);                           /* input size               */
    for( j=0,i=1; i<=MAX_K_SIZE; i=i*2 ){     /* check user input         */
      if( i==k ){
	j++;           /* set j=1 if k = power of 2 that is <= MAX_K_SIZE */
	break;
      }
    }
  } while ( j==0 );
  mem_size = 1024 * k; /* return size in bytes   */
  k_size = k;          /* return size in K bytes */

                       /* input programming period (or leave unchanged) */
  printf("\n\
Select programming time period and device type: 10ms=EEprom; 50ms=ePROM.\n\
Press 1 for EEprom or 5 for ePROM ? ");
  switch( getchar() ){
  case '5': tdel2 = tdel2_50; chip_type = EPROM; break;
  case '1': tdel2 = tdel2_10; chip_type = EEPROM; break;
  }
}

/*************************************************************************/

/* define on line help messages */

char *epc_help[] = { "\n\n\n\n",
  "- HELP ---------------------------------------------------------------\n",
  "* Data can be moved between DISK and buffers A or B,  and between PROM\n",
  "  and buffers A or B.  The  command  diagram shows the default buffers\n",
  "  used by commands to transfer data but any buffer may be specified.\n",
  "* A typical programming session would involve the commands:  L to load\n",
  "  in HEX data into buffer A; R to read the PROM and check erasure; and\n",
  "  P to program buffer A into the PROM (with automatic PROM check).\n",
  "* Formatted HEX data can be read in three formats via command L:\n",
  "  1 Motorola \"S1\" records, 2 Tektronix HEX, or 3 INTEL HEX (type 00).\n",
  "* Unformatted HEX data can be read via command C-L (control_L).\n",
  "* Disk files are written in Motorola \"S1\" record format only.\n",
  "." };

char *misc_help[] = { "\n Additional user commands:\n",
" #   show internal data decode lookup table (used in input port decode).\n",
" @   calibrate timer constants (useful in development).\n",
" !   work through hardware tests (read part 2 of the project description).\n",
" ?   displays software/author details.\n",
" C-l (press control_L) load buffer from file of unformatted HEX characters\n",
"     where the file contains a stream of HEX characters {0..9,a..f,A..F}.\n",
"     Formatting characters such as new_line, tab, etc are ignored so long\n",
"     as there are NO MORE THAN 2 CONSECUTIVE non-hex characters.\n",
"\n",
" To suppress the exit message, run the program with the \"-smith\" switch.\n",
"\n",
" To do: 1) Add intelligent programming algorithms.\n",
"        2) Possibly add more output file formats.\n",
"\n",
"." };

/*************************************************************************/

/*
 * buffer_copy() copies data between buffers.
 */
void buffer_copy()
{
  printf("opy buffers  A x B  where x in {<,>} ? ");
  switch( getchar() ){
  case '<':
  case ',': for( addr=0; addr<mem_size; addr++ )
    a[addr] = b[addr];
    printf("\nCopied A <-- B\n");
    break;
  case '>':
  case '.': for( addr=0; addr<mem_size; addr++ )
    b[addr] = a[addr];
    printf("\nCopied A --> B\n");
    break;
  default:  printf("\nEnter < to copy B to A and > to copy A to B\n");
    break;
  }
}

/*************************************************************************/

/*
 * buffer_test() compares buffer data and notes if buffer == $FF (empty).
 */
void buffer_test()
{
  printf("est buffers A&B - ");
  /* compare bytes between buffers, note 0xFF (empty) */
  for( addr=i=j=k=0; addr<mem_size; addr++ ){
    if( a[ addr ] != b[ addr ] ) i++;
    if( a[ addr ] != EMPTY ) j++;
    if( b[ addr ] != EMPTY ) k++;
  }
  printf("%d bytes differ in %d total.\n",i,mem_size);
  if( j == 0 ){
    printf("Buffer A is empty.\n");
  } else {
    printf("Buffer A holds %d bytes <> $FF.\n",j);
  }
  if( k == 0 ){
    printf("Buffer B is empty.\n");
  } else {
    printf("Buffer B holds %d bytes <> $FF.\n",k);
  }
}

/*************************************************************************/

/*
 * HEX input and output functions, and associated global variables.
 */

FILE *hex_file;         /* pointer to HEX file     */

int chk_in, chk_out;    /* checksums               */

char line[256];         /* buffer of 1 line of HEX */
int nc,ip;              /* total chars read, index of current char to read */
int sbc,nbytes;

/*
 * get_byte() returns  the value of the  next 2 hex chars  in line[].  Byte
 * checksum chk_in is updated.
 */
int get_byte()
{
  int ans;

  sscanf( &line[ip],"%2x",&ans );      /* read value of 2 HEX characters */
  ip = ip+2;                           /* advance read index by 2        */

  chk_in = (chk_in + ans) & MASK8;     /* update 8bit checksum           */
  return ans;
}

/*
 * get_byte2() returns the value of the next 2 hex chars in line[].  Nibble
 * checksum chk_in is updated.
 */
int get_byte2()
{
  int n1,n2;

  sscanf( &line[ip],"%1x%1x",&n1,&n2 );/* read values of two HEX chars */
  ip = ip+2;                           /* advance read index by 2      */

  chk_in = (chk_in + n1 + n2) & MASK8; /* update 4bit checksum         */
  return (n1 << 4 | n2);
}

/*
 * out_hex(n) outputs n as 2 hex chars.  Byte checksum chk_out is updated.
 */
void out_hex( n )
int n;
{
  int x;

  x = n & MASK8;
  fprintf(hex_file,"%02x",x);

  chk_out = (chk_out + x) & MASK8;
}

/*
 * bload(buf) loads  buffer buf with data from  a disk file.  Note that the
 * descriptive  strings accessed  via  describe_hex[] are  ordered to match
 * symbols hexmotorola, hexintel, hextek, etc.
 */

#define hexmotorola 0
#define hexintel 1
#define hextek 2
char *describe_hex[] = { "motorola S-records",
                         "intel",
                         "tek" };

void bload( buf )
char* buf;
{
  int addr_begin;
  int hex_type;

  printf("\n  Input file? ");
  scanf("%s",line);
  if( (hex_file = fopen( line,"r" )) == NULL ){
    printf("? can't open input file: %s\n",line);
    return;
  }

  printf("\n");
  k = 0;                               /* clear input byte counter */
  hex_type = -1;

                                       /* each line of text in the input file
                                        * is read into line[] and decoded.
                                        */

  while( (nc=fgets(line, 256, hex_file) != NULL) ){
    for( ip=0;
         (ip<nc) & ( (line[ip]!='S')&(line[ip]!=':')&(line[ip]!='/') );
          ip++ );                      /* search for { S : / } character    */
    if( ip == nc ) break;              /* skip input line if no S or : or / */

/* - - - - - - - - - - - - - - - - - - */
    if( line[ip] == 'S' ){             /* read motorola S records */
      hex_type = hexmotorola;

      ip++;
      if (line[ip] == '9') break;      /* exit on S9 */
      if (line[ip] == '1'){            /* process S1 data record   */
        ip++;
        chk_in = 0;                    /* clear input checksum     */
        sbc = get_byte();              /* input record's bytecount */
        addr = 256*get_byte()+get_byte();      /* input address    */
        addr_begin = addr;
        for( ; addr >= mem_size ; addr = addr - mem_size );
        printf("\r%04x: ",addr);

        nbytes = sbc-3;                /* =record's DATA bytecount */
        for( i=0; i<nbytes; i++ ){     /* input data               */
          buf[ addr ] = ( j = get_byte() );
          addr++; k++;
          if( addr == mem_size ) addr = 0;
          printf("%02x",j);
        }
        i = get_byte();                /* input checksum, update chk_in */

        if( ( (1+chk_in) & MASK8 ) == 0 ){
          printf(" ok");
        } else {
          printf("\n? MOTOROLA checksum error in record at $%04x\n",addr_begin );
          return;
        }
      }                                /* end of handling of S record */

/* - - - - - - - - - - - - - - - - - - */
    } else if( line[ip] == ':' ){      /* handle intel hex */
      hex_type = hexintel;

      ip++;
      chk_in = 0;                      /* clear input checksum          */
      nbytes = get_byte();             /* input record's DATA bytecount */
      addr = 256*get_byte()+get_byte();/* input address                 */
      addr_begin = addr;
      for( ; addr >= mem_size ; addr = addr - mem_size );
      printf("\r%04x: ",addr);

      if( get_byte() == 0 ){           /* currently handle INTEL type 00 */
        for( i=0; i<nbytes; i++ ){     /* input data                     */
          buf[ addr ] = ( j = get_byte() );
          addr++; k++;
          if( addr == mem_size ) addr = 0;
          printf("%02x",j);
        }

        if( (~(chk_in - 1) & MASK8) == get_byte() ){   /* check checksum */
          printf(" ok");
        } else {
          printf("\n? INTEL checksum error in record at $%04x\n",addr_begin );
          return;
        }
      }
       /* finished decoding this INTEL hex record */

/* - - - - - - - - - - - - - - - - - - */
    } else if( line[ip] == '/' ){      /* handle tek hex */
      hex_type = hextek;

      ip++;
      chk_in = 0;                              /* clear input checksum  */
      addr = 256*get_byte2()+get_byte2();      /* input address         */
      addr_begin = addr;
      nbytes = get_byte2();            /* input record's DATA bytecount */
      i = chk_in;
      if( i != (j=get_byte2()) ){
        printf("\n? checksum error in TEK *header* at $%04x\n",addr_begin );
        return;
      }

      if( nbytes == 0 ) break;
      for( ; addr >= mem_size ; addr = addr - mem_size );
      printf("\r%04x: ",addr);

      chk_in = 0;                      /* clear check sum for DATA read */
      for( i=0; i<nbytes; i++ ){       /* input data */
        buf[ addr ] = ( j = get_byte2() );
        addr++; k++;
        if( addr == mem_size ) addr = 0;
        printf("%02x",j);
      }

      i = chk_in;
      if( i == get_byte2() ){
        printf(" ok");
      } else {
        printf("\n? TEK checksum error in record at $%04x\n",addr_begin );
        return;
      }
      /* finished decoding this TEK hex record */

/* - - - - - - - - - - - - - - - - - - */
    }                          /* end of decoding HEX record */

  }                            /* finished processing file */

  fclose( hex_file );
  printf("\nBytes read: %d of %s HEX data\n",k,describe_hex[ hex_type ]);
}

/*
 * hload(buf) loads buffer buf from a plain file of HEX characters.
 */

void hload( buf )
char* buf;
{
  printf("\n  Input file? ");
  scanf("%s",line);
  if( (hex_file = fopen( line,"r" )) == NULL ){
    printf("? can't open input file: %s\n",line);
    return;
  }

  printf("\n Address Offset (4 HEX characters) ? ");
  scanf( "%4x",&addr );
  if( addr > mem_size ){
    printf("\n? Address Offset too large, 0 used.\n");
      addr = 0;
  }
  printf("\n");

  k = 0;                                  /* clear input byte counter */
  for( ; ; ){
           /* assume maximum number of consecutive non-HEX chars is 2 */
    if( (i = fgetc( hex_file )) == FGETC_EOF ) break; /* input char 1 */
    if( i < '0' ) i = fgetc( hex_file );
    if( i == FGETC_EOF ) break;
    if( i < '0' ) i = fgetc( hex_file );
    if( i == FGETC_EOF ) break;

    if( (j = fgetc( hex_file )) == FGETC_EOF ) break; /* input char 2 */
    if( j < '0' ) j = fgetc( hex_file );
    if( j == FGETC_EOF || j < '0' ) break;

    line[0] = (char) i; line[1] = (char) j; line[2] = '\0';

    if( (addr & 0x1f) == 0 ) printf("\r%04x:",addr);
    if( (addr & 7) == 0 ) printf(" ");

    ip = 0; buf[ addr ] = get_byte();  /* decode two HEX characters and store */
    printf("%02x", buf[ addr ] );
    addr++; k++;
    if( addr == mem_size ) addr = 0;
  }

  fclose( hex_file );
  printf("\nBytes read: %d of HEX\n",k);
}

/*
 * buf_write(buf) writes data in  buffer buf to an  S-record HEX data file.
 * An address offset may be applied.
 */

void buf_write( buf )
     char* buf;
{
  unsigned addr_offset;

  printf("\n Output file? ");
  scanf("%s",line);
  if( (hex_file = fopen( line,"w" )) == NULL ){
    printf("? can't open output file: %s\n",line);
    return;
  }

  printf("\n Address Offset (4 HEX characters) ? ");
  scanf( "%4x",&addr_offset );

  printf("\n");
  for( addr=0; addr<mem_size; ){
    fprintf(hex_file,"S1");
    chk_out = 0;                       /* clear checksums */

    nbytes = mem_size - addr;
    if( nbytes > BPL ) nbytes = BPL;
    out_hex( nbytes+3 );               /* total bytes in record */

    i = addr_offset + addr;            /* get effective address */
    printf("\r%04x:",i);
    out_hex( (i >> 8) & MASK8 );
    out_hex( i & MASK8 );

    for( i=0; i<nbytes; i++ ){         /* output nbytes data bytes  */
      if( (addr & 7) == 0 ) printf(" ");
      out_hex( buf[ addr ] );
      printf("%02x", buf[ addr ] );
      addr++;
    }
    out_hex( (-1-chk_out) & MASK8 );   /* output checksum */
    fprintf(hex_file,"\n");
  }
  fprintf(hex_file,"S9");
  chk_out = 0; out_hex( 3 ); out_hex( 0 ); out_hex( 0 );
  out_hex( (-1-chk_out) & MASK8 );
  fprintf(hex_file,"\n");

  fclose( hex_file );
  printf("\nBytes written: %d\n",mem_size);
}

/*************************************************************************/

/*
 * buf_empty(buf) sets buffer to empty.
 */

void buf_empty( buf )
     char* buf;
{
  for( addr=0; addr<mem_size; addr++ ){
    buf[ addr ] = EMPTY;
    }
  printf("set to EMPTY\n");
}

/*************************************************************************/

/*
 * buf_view(buf) displays buffer buf.
 */

void buf_view( buf )
     char* buf;
{
  char cdata[32];

  printf("\n");
  k = 0;
  for( addr=0; addr<mem_size; ){
    nbytes = mem_size - addr;
    if( nbytes > 16 ) nbytes = 16;
    printf("\r%04x: ",addr);
    for( i=j=0; i<nbytes; i++ ){       /* output nbytes data bytes  */
      printf("%02x", buf[ addr ] );
      cdata[ i ] = buf[ addr ] & 0x7f;
      if( cdata[ i ] < ' ' || cdata[ i ] > 0x7e ) cdata[ i ] = ' ';
      if( buf[ addr ] != EMPTY ) j++;
      addr++;
    }
    cdata[ i ] = '\0';
    printf( " |%s|",cdata );

    if( j > 0 ){
      if( ++k == 24 ){
        k = 0;
        printf(" q=quit, \' \' for more ? ");
        if( getchar() == 'q' ) return;
      }
      printf("\n");
    }
  }
  printf(" Press any key to exit ");
  getchar();
  printf("\n");
}

/*************************************************************************/

/*
 * eread(buf) reads contents of PROM into buffer buf.
 */

void eread( buf )
     char* buf;
{
  unsigned addr_offset;

  _outb( 0x05, CTRL );                  /* select READ mode      */
  _outb( 0x08, CTRL );                  /* clear address counter */
  _outb( 0x05, CTRL );

  printf("\n Programmer address cleared for read.  Pre-increment (n/y) ? ");
  switch( getchar() ){
  case 'y':
  case 'Y': printf("\n Address Offset (4 HEX characters) ? ");
            scanf( "%4x",&addr_offset );
            for( i=0; i<addr_offset; i++ ){
              _outb( 0x04, CTRL );                /* inc address counter */
              _outb( 0x05, CTRL );
            }
            break;
  }

  printf("\n");
  for( addr=0; addr<mem_size; addr++ ){
    if( (addr & 31) == 0 ){
      printf("\r%04x:",addr);
    }
    if( (addr & 7) == 0 ) printf(" ");

    printf( "%02x", ( *(buf + addr) = read_byte() ) );

    _outb( 0x04, CTRL );                /* inc address counter */
    _outb( 0x05, CTRL );
  }
  printf("\n");

  for( i=addr=0; addr<mem_size; addr++ )        /* check for erasure */
    if( *(buf + addr) != EMPTY ) i++;
  if ( i == 0 ) printf("Device is erased.\n");
}

/*************************************************************************/

/*
 * ewrite(buf) writes buf into the (E)EPROM.
 */

void ewrite( buf )
     char* buf;
{
  printf("\n\
Insert erased (E)EPROM, enable power, select correct Vpp, & press G to go... ");
  if( getchar() != 'G' ) return;
  printf("\n");

  _outb( 0x05, CTRL );                  /* select READ mode      */
  _outb( 0x08, CTRL );                  /* clear address counter */
  _outb( 0x05, CTRL );

  k = 0;                                /* clear prog. fail counter */
  for( addr=0; addr<mem_size; addr++ ){
    if( (addr & 31) == 0 ){
      printf("\r%04x:",addr);
    }
    if( (addr & 7) == 0 ) printf(" ");

               /* write DATA if device==EEprom or value <> EMPTY */
    if( (chip_type == EEPROM) || (*(buf + addr) != EMPTY) ){
      _outb( 0x0b, CTRL );              /* turn on Vpp           */
      _outb( *(buf + addr), DOUT );     /* send data byte        */

      _outb( 0x03, CTRL );              /* start WRITE           */
      time_delay( tdel1,tdel2 );        /* wait ...              */
      _outb( 0x05, CTRL );              /* stop WRITE, read back */

      if( read_byte() != *(buf + addr) ){
        printf( "??" );
        k++;
      } else {
        printf( "%02x", *(buf + addr) );
      }
    } else {           /* note that this EMPTY location was not programmed */
      printf( "__" );  /* (we assume the EPROM is EMPTY || user is smart)  */
    }

    _outb( 0x04, CTRL );                /* inc address counter */
    _outb( 0x05, CTRL );
  }
  printf("\n");

  if ( k == 0 ){
    printf("All bytes programmed correctly.\n");
  } else {
    printf("? %d bytes not programmed correctly.\n",k);
  }
}

/*************************************************************************/

/*
 * hardware_check() provides  a basic test  of cables  and basic functions,
 * roughly as described in the design article part 2.
 */

void hardware_check()
{
  printf("\nTest mode\n\nREMOVE PROMs then press a key to proceed. ");
  getchar();
  printf("\n");

  /* * * * * * * * * * * * * * * * * * * * * * * */

  _outb( 0x00, DOUT );                  /* zero data out lines   */
  _outb( 0x08, CTRL );
  _outb( 0x03, CTRL );
  printf("All pins => zero volts except SOC2_p1, SOC2_p22, and Vcc\n");
  printf("Press any key to continue ...");
  getchar();

  printf("\n\nNow proceeding to test individual data bus bits\n");
  for( i=1,j=0; i!=256; j++ ){
    _outb( i, DOUT ); 
    printf("\ndata bit %d set (%d), press any key to continue",j,i);
    getchar();
    i = i<<1;
  }

  /* * * * * * * * * * * * * * * * * * * * * * * */

  printf("\n\nNow proceeding to test individual address bus bits\n");
  _outb( 0x05, CTRL );                  /* select READ mode      */
  _outb( 0x08, CTRL );                  /* clear address counter */
  _outb( 0x05, CTRL );
  for( addr=0,i=j=1; i<14; i++ ){
    for( ; addr < j; addr++ ){
      _outb( 0x04, CTRL );              /* inc address counter   */
      _outb( 0x05, CTRL );
    }
    printf("\naddress bit %d set, press any key to continue",i-1);
    getchar();
    j=j<<1;
  }

  /* * * * * * * * * * * * * * * * * * * * * * * */

  printf("\n\n\
We can now verify data read capability by connecting A0-A7 to D0-D7.\
If you wish to conduct this test, enter y ? ");
  if( getchar() != 'y' ){
    printf("\nEnd of test.\n");
    return;
  }

  printf("\n");
  _outb( 0x05, CTRL );                  /* select READ mode      */
  _outb( 0x08, CTRL );                  /* clear address counter */
  _outb( 0x05, CTRL );
  for( addr=0,i=j=1; i<8; i++ ){
    for( ; addr < j; addr++ ){
      _outb( 0x04, CTRL );              /* inc address counter   */
      _outb( 0x05, CTRL );
    }
    if( read_byte() != addr ){
      printf("? mismatch on bit %d\n",i-1);
    } else {
      printf("Bit %d is OK\n",i-1);
    }
    j=j<<1;
  }
  printf("\n");
}

/*************************************************************************/

/*
 * show_lookup_table() displays the data read decode table.  Humans have no
 * place here.
 */

void show_lookup_table()
{
  k=0;
  for( i=0; i<=15; i++ ) {
    for( j=0; j<=15; j++ ) {
      printf(" %d %d\n",k,lookup[k]);
      k++;
    };
    getchar();
  }
}

/*************************************************************************/

/*
 * main() performs some system initialisation before  entering a loop which
 * handles user commands.
 */

char *prompt[]={ "\n   - %s --------------------------- %3dK %s @ %3d ms ----",
      "\n  |                            BUFFERS                             |",
      "\n  |    DISK   -------Load---->  [ A ]   ----Program----->  PROM    |",
      "\n  |    FILE  <------Write----   [ B ]  <----Read--------   PROG.   |",
      "\n  |                                                                |",
      "\n  | l)oad file into buffer A/B    | t)est buffers A versus B       |",
      "\n  | r)ead PROM into buffer B/A    | v)iew buffer B/A               |",
      "\n  | p)rogram buffer A/B into PROM | c)opy buffer to buffer         |",
      "\n  | w)rite buffer B/A to file     | x) exchange buffers A and B    |",
      "\n  | q)uit   h)elp   m)isc_help    | s)elect device | e)mpty buffer |",
      "\n   ----------------------------------------------------------------",
      "\n  Command? ",
      "." };

void main(argc,argv)
int  argc;
char *argv[];
{
                       /* initialise the data decode lookup table       */
  for( i=0; i<256; i++ )
    lookup[i] = 999;   /* first fill ALL elements with an invalid value */

                       /* then fill 16 locations with valid 4bit data   */
  for( j=0; j<=15; j++ ) {
    addr = (((unsigned)j & 1) ^ 1) << 7 | 
      ((unsigned)j & 2) << 4 | 
        ((unsigned)j & 4) << 2 | 
          ((unsigned)j & 8);
    lookup[ addr+7 ] = j;
  }                    /* so lookup[8bit_port_data] ==> valid nibble!   */

  print_array( pname );                               /* output welcome */
  printf( "\n%s\n",themag );  

  for( i=0; i<MAXMS; i++ )     /* clear buffers with $FF   */
    a[ i ] = b[ i ] = EMPTY;

  tdel2_10 = 20;               /* assign "internal" time delay loop counts */
  tdel2_50 = 100;              /* (do this BEFORE select_device() called)  */
  calibrate_timer( 0 );        /* calibrate time keeping                   */

  chip_type = EPROM;           /* default chip type is EPROM */
  select_device();             /* choose initial PROM size and timing */

  print_array( epc_help );
  do {                         /* enter main command loop */

    _outb( 0x05, CTRL );       /* select READ mode        */
    _outb( 0x08, CTRL );       /* clear address counter   */
    _outb( 0x05, CTRL );

    times( now );              /* prompt for command */
    printf( prompt[0],now,k_size,device_name[ chip_type ],tdel2/2 );
    for( i=1; *prompt[i] != '.'; i++ )
      printf( prompt[i] );

    switch( cmd = getchar() ){ /* read and decode command */

    case 'c':                  /* copy data between buffers */
    case 'C': buffer_copy();
      break;

    case 'e':                  /* set buffer to empty */
    case 'E': printf("mpty buffer ");
      bmsg = bmsg_a;
      buf_empty( buffer() );
      break;

    case 'h':                  /* display help */
    case 'H': print_array( epc_help );
      break;

    case 'l':                  /* load data from disk into a buffer */
    case 'L': printf("oad buffer ");
      bmsg = bmsg_a;
      bload( buffer() );
      break;

    case 'l'-0x40:             /* load data from hex dump file */
    case 'L'-0x40: printf("\n Load from a HEX file dump into buffer ");
      bmsg = bmsg_a;
      hload( buffer() );
      break;

    case 'p':                  /* write buffer data into PROM */
    case 'P': printf("rogram buffer ");
      bmsg = bmsg_a;
      ewrite( buffer() );
      break;

    case 'q':                  /* quit from program */
    case 'Q': cmd = 'q';
      printf("uit\n");
      break;

    case 'r':                  /* copy data from PROM into buffer */
    case 'R': printf("ead buffer ");
      bmsg = bmsg_b;
      eread( buffer() );
      break;

    case 's':                  /* select maximum memory size */
    case 'S': select_device(); /* and the programming period */
      break;

    case 't':
    case 'T': buffer_test();
      break;

    case 'v':                  /* view buffer contents */
    case 'V': printf("iew buffer ");
      bmsg = bmsg_b;
      buf_view( buffer() );
      break;

    case 'w':                  /* write buffer to disk HEX file */
    case 'W': printf("rite buffer ");
      bmsg = bmsg_b;
      buf_write( buffer() );
      break;

    case 'x':                  /* eXchange buffer contents */
    case 'X': for( i=0; i<MAXMS; i++ ){
      c = a[ i ];
      a[ i ] = b[ i ];
      b[ i ] = c;
    }
      printf(" - Buffers A and B exchanged.\n");
      break;

    case '#': show_lookup_table();     /* test 1 -pm */
      break;

    case '@': printf("\n");            /* test 2 -pm */
      calibrate_timer( 1 );
      break;

    case '!': hardware_check();        /* test 3 - construction test */
      break;

    case '?':                          /* output whoami */
      print_array( pname );
      printf("\n %s %s\n",whoami,email);
      printf("\n%s\n%s\n",themag,theact);
      break;

    case 'm':
    case 'M':                          /* display miscellaneous help */
      print_array( misc_help );
      printf("\n Press a key to continue ... ");
      getchar();
      break;

    }
  } while ( cmd != 'q' );

  if( !( argc > 0 && *argv[1] == '-' && *(argv[1]+1) == 's' ) ){
    printf("%s",smith);
  }
}

/*************************************************************************/
